#include "squinewave.h"

namespace squinewave {

/*******************************************************************************************************************
Cycling '74 License for Max-Generated Code for Export
Copyright (c) 2016 Cycling '74
The code that Max generates automatically and that end users are capable of exporting and using, and any
  associated documentation files (the “Software”) is a work of authorship for which Cycling '74 is the author
  and owner for copyright purposes.  A license is hereby granted, free of charge, to any person obtaining a
  copy of the Software (“Licensee”) to use, copy, modify, merge, publish, and distribute copies of the Software,
  and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The Software is licensed to Licensee only for non-commercial use. Users who wish to make commercial use of the
  Software must contact the copyright owner to determine if a license for commercial use is available, and the
  terms and conditions for same, which may include fees or royalties. For commercial use, please send inquiries
  to licensing (at) cycling74.com.  The determination of whether a use is commercial use or non-commercial use is based
  upon the use, not the user. The Software may be used by individuals, institutions, governments, corporations, or
  other business whether for-profit or non-profit so long as the use itself is not a commercialization of the
  materials or a use that generates or is intended to generate income, revenue, sales or profit.
The above copyright notice and this license shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
  THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
*******************************************************************************************************************/

// global noise generator
Noise noise;
static const int GENLIB_LOOPCOUNT_BAIL = 100000;


// The State struct contains all the state and procedures for the gendsp kernel
typedef struct State {
	CommonState __commonstate;
	Delta __m_delta_7;
	int vectorsize;
	int __exception;
	t_sample m_sq_1;
	t_sample m_param_smoothing_4;
	t_sample samplerate;
	t_sample m_freq_3;
	t_sample m_freqlimit_5;
	t_sample m_w_2;
	// re-initialize all member variables;
	inline void reset(t_param __sr, int __vs) {
		__exception = 0;
		vectorsize = __vs;
		samplerate = __sr;
		m_sq_1 = ((int)0);
		m_w_2 = ((int)0);
		m_freq_3 = ((int)0);
		m_param_smoothing_4 = ((t_sample)0.01);
		m_freqlimit_5 = ((int)2205);
		__m_delta_7.reset(0);
		genlib_reset_complete(this);
		
	};
	// the signal processing routine;
	inline int perform(t_sample ** __ins, t_sample ** __outs, int __n) {
		vectorsize = __n;
		const t_sample * __in1 = __ins[0];
		const t_sample * __in2 = __ins[1];
		const t_sample * __in3 = __ins[2];
		t_sample * __out1 = __outs[0];
		t_sample * __out2 = __outs[1];
		if (__exception) {
			return __exception;
			
		} else if (( (__in1 == 0) || (__in2 == 0) || (__in3 == 0) || (__out1 == 0) || (__out2 == 0) )) {
			__exception = GENLIB_ERR_NULL_BUFFER;
			return __exception;
			
		};
		int min_6 = (-1);
		// the main sample loop;
		while ((__n--)) {
			const t_sample in1 = (*(__in1++));
			const t_sample in2 = (*(__in2++));
			const t_sample in3 = (*(__in3++));
			t_sample clamp_52 = ((in3 <= min_6) ? min_6 : ((in3 >= ((int)1)) ? ((int)1) : in3));
			t_sample mul_51 = (clamp_52 * ((t_sample)0.5));
			t_sample rsub_50 = (((t_sample)0.5) - mul_51);
			t_sample mix_71 = (m_w_2 + (m_param_smoothing_4 * (rsub_50 - m_w_2)));
			t_sample mix_53 = mix_71;
			t_sample clamp_56 = ((in2 <= ((int)0)) ? ((int)0) : ((in2 >= ((int)1)) ? ((int)1) : in2));
			t_sample rsub_55 = (((t_sample)1.00001) - clamp_56);
			t_sample rdiv_54 = safediv(((int)1), rsub_55);
			t_sample mix_72 = (m_sq_1 + (m_param_smoothing_4 * (rdiv_54 - m_sq_1)));
			t_sample mix_57 = mix_72;
			t_sample div_49 = safediv(mix_53, mix_57);
			t_sample rsub_47 = (((int)1) - mix_53);
			t_sample div_48 = safediv(rsub_47, mix_57);
			t_sample delta_46 = __m_delta_7(in1);
			t_sample wrap_45 = wrap(delta_46, (-0.5), ((t_sample)0.5));
			t_sample mul_44 = (wrap_45 * samplerate);
			t_sample abs_43 = fabs(mul_44);
			int gt_42 = (abs_43 > m_freq_3);
			t_sample mix_73 = (abs_43 + (((t_sample)0.999) * (m_freq_3 - abs_43)));
			t_sample switch_41 = (gt_42 ? abs_43 : mix_73);
			t_sample div_39 = safediv(switch_41, m_freqlimit_5);
			t_sample mul_38 = (div_39 * ((t_sample)0.5));
			t_sample clamp_37 = ((mul_38 <= ((int)0)) ? ((int)0) : ((mul_38 >= ((t_sample)0.5)) ? ((t_sample)0.5) : mul_38));
			t_sample rsub_36 = (((int)1) - clamp_37);
			t_sample clamp_35 = ((div_49 <= clamp_37) ? clamp_37 : ((div_49 >= rsub_36) ? rsub_36 : div_49));
			t_sample div_31 = safediv(in1, clamp_35);
			t_sample clamp_28 = ((div_31 <= ((int)0)) ? ((int)0) : ((div_31 >= ((int)1)) ? ((int)1) : div_31));
			t_sample clamp_34 = ((div_48 <= clamp_37) ? clamp_37 : ((div_48 >= rsub_36) ? rsub_36 : div_48));
			t_sample rsub_32 = (((int)1) - clamp_34);
			t_sample clamp_33 = ((mix_53 <= clamp_35) ? clamp_35 : ((mix_53 >= rsub_32) ? rsub_32 : mix_53));
			int lte_25 = (in1 <= clamp_33);
			t_sample sub_29 = (in1 - clamp_33);
			t_sample div_30 = safediv(sub_29, clamp_34);
			t_sample clamp_27 = ((div_30 <= ((int)0)) ? ((int)0) : ((div_30 >= ((int)1)) ? ((int)1) : div_30));
			t_sample rsub_26 = (((int)1) - clamp_27);
			t_sample switch_24 = (lte_25 ? clamp_28 : rsub_26);
			t_sample out2 = switch_24;
			t_sample mul_58 = (switch_24 * ((t_sample)3.1415926535898));
			t_sample cos_59 = cos(mul_58);
			t_sample out1 = cos_59;
			t_sample w_next_60 = fixdenorm(mix_53);
			t_sample sq_next_61 = fixdenorm(mix_57);
			m_w_2 = w_next_60;
			m_sq_1 = sq_next_61;
			// assign results to output buffer;
			(*(__out1++)) = out1;
			(*(__out2++)) = out2;
			
		};
		return __exception;
		
	};
	inline void set_param_smoothing(t_param _value) {
		m_param_smoothing_4 = (_value < 0 ? 0 : (_value > 1 ? 1 : _value));
	};
	inline void set_freqlimit(t_param _value) {
		m_freqlimit_5 = (_value < 1e-05 ? 1e-05 : (_value > 11025 ? 11025 : _value));
	};
	
} State;


///
///	Configuration for the genlib API
///

/// Number of signal inputs and outputs

int gen_kernel_numins = 3;
int gen_kernel_numouts = 2;

int num_inputs() { return gen_kernel_numins; }
int num_outputs() { return gen_kernel_numouts; }
int num_params() { return 2; }

/// Assistive lables for the signal inputs and outputs

const char *gen_kernel_innames[] = { "phasor:unipolar", "squareness:unipolar", "skew:bipolar" };
const char *gen_kernel_outnames[] = { "squinewave:audio", "ramp:unipolar" };

/// Invoke the signal process of a State object

int perform(CommonState *cself, t_sample **ins, long numins, t_sample **outs, long numouts, long n) {
	State* self = (State *)cself;
	return self->perform(ins, outs, n);
}

/// Reset all parameters and stateful operators of a State object

void reset(CommonState *cself) {
	State* self = (State *)cself;
	self->reset(cself->sr, cself->vs);
}

/// Set a parameter of a State object

void setparameter(CommonState *cself, long index, t_param value, void *ref) {
	State *self = (State *)cself;
	switch (index) {
		case 0: self->set_freqlimit(value); break;
		case 1: self->set_param_smoothing(value); break;
		
		default: break;
	}
}

/// Get the value of a parameter of a State object

void getparameter(CommonState *cself, long index, t_param *value) {
	State *self = (State *)cself;
	switch (index) {
		case 0: *value = self->m_freqlimit_5; break;
		case 1: *value = self->m_param_smoothing_4; break;
		
		default: break;
	}
}

/// Get the name of a parameter of a State object

const char *getparametername(CommonState *cself, long index) {
	if (index >= 0 && index < cself->numparams) {
		return cself->params[index].name;
	}
	return 0;
}

/// Get the minimum value of a parameter of a State object

t_param getparametermin(CommonState *cself, long index) {
	if (index >= 0 && index < cself->numparams) {
		return cself->params[index].outputmin;
	}
	return 0;
}

/// Get the maximum value of a parameter of a State object

t_param getparametermax(CommonState *cself, long index) {
	if (index >= 0 && index < cself->numparams) {
		return cself->params[index].outputmax;
	}
	return 0;
}

/// Get parameter of a State object has a minimum and maximum value

char getparameterhasminmax(CommonState *cself, long index) {
	if (index >= 0 && index < cself->numparams) {
		return cself->params[index].hasminmax;
	}
	return 0;
}

/// Get the units of a parameter of a State object

const char *getparameterunits(CommonState *cself, long index) {
	if (index >= 0 && index < cself->numparams) {
		return cself->params[index].units;
	}
	return 0;
}

/// Get the size of the state of all parameters of a State object

size_t getstatesize(CommonState *cself) {
	return genlib_getstatesize(cself, &getparameter);
}

/// Get the state of all parameters of a State object

short getstate(CommonState *cself, char *state) {
	return genlib_getstate(cself, state, &getparameter);
}

/// set the state of all parameters of a State object

short setstate(CommonState *cself, const char *state) {
	return genlib_setstate(cself, state, &setparameter);
}

/// Allocate and configure a new State object and it's internal CommonState:

void *create(t_param sr, long vs) {
	State *self = new State;
	self->reset(sr, vs);
	ParamInfo *pi;
	self->__commonstate.inputnames = gen_kernel_innames;
	self->__commonstate.outputnames = gen_kernel_outnames;
	self->__commonstate.numins = gen_kernel_numins;
	self->__commonstate.numouts = gen_kernel_numouts;
	self->__commonstate.sr = sr;
	self->__commonstate.vs = vs;
	self->__commonstate.params = (ParamInfo *)genlib_sysmem_newptr(2 * sizeof(ParamInfo));
	self->__commonstate.numparams = 2;
	// initialize parameter 0 ("m_freqlimit_5")
	pi = self->__commonstate.params + 0;
	pi->name = "freqlimit";
	pi->paramtype = GENLIB_PARAMTYPE_FLOAT;
	pi->defaultvalue = self->m_freqlimit_5;
	pi->defaultref = 0;
	pi->hasinputminmax = false;
	pi->inputmin = 0;
	pi->inputmax = 1;
	pi->hasminmax = true;
	pi->outputmin = 1e-05;
	pi->outputmax = 11025;
	pi->exp = 0;
	pi->units = "";		// no units defined
	// initialize parameter 1 ("m_param_smoothing_4")
	pi = self->__commonstate.params + 1;
	pi->name = "param_smoothing";
	pi->paramtype = GENLIB_PARAMTYPE_FLOAT;
	pi->defaultvalue = self->m_param_smoothing_4;
	pi->defaultref = 0;
	pi->hasinputminmax = false;
	pi->inputmin = 0;
	pi->inputmax = 1;
	pi->hasminmax = true;
	pi->outputmin = 0;
	pi->outputmax = 1;
	pi->exp = 0;
	pi->units = "";		// no units defined
	
	return self;
}

/// Release all resources and memory used by a State object:

void destroy(CommonState *cself) {
	State *self = (State *)cself;
	genlib_sysmem_freeptr(cself->params);
		
	delete self;
}


} // squinewave::
